// Java Interface to Postgres
// $Id: PGConnection.java,v 1.2 1997/04/12 21:30:40 finder Exp $

// Copyright (c) 1996 Bradley McLean / Jeffrey Medeiros
// Modifications Copyright (c) 1996/1997 Martin Rode
// Copyright (c) 1997 Peter T Mount
//
// You may distribute under the terms of the GNU Public License as
// specified in the README file that comes with the JavaPostgres95 Kit

package postgres95;

import java.sql.*;
import java.math.*;
import java.net.*;
import java.io.*;

/**
 * JDBC Interface to Postgres95 functions
 */

public class PGConnection implements Connection
{
   PGDriver parent;

   String Host;
   int Port;
   String Options;
   String Tty;
   String Dbname;
   String User;
   String Pass;

   DataInputStream sock_in;
   DataOutputStream sock_out;

   boolean autoCommit = true;
   boolean ReadOnly = false;
   String Catalog = null;
   static final int TRANSACTION_NONE		 = 0;
   static final int TRANSACTION_READ_UNCOMMITTED = 1;
   static final int TRANSACTION_READ_COMMITTED   = 2;
   static final int TRANSACTION_REPEATABLE_READ  = 3;
   static final int TRANSACTION_SERIALIZABLE     = 4;
   static final int TRANSACTION_VERSIONING       = 5;
   int transactionIsolation = TRANSACTION_NONE;
   boolean autoClose = true;
   SQLWarning warning = null;

   private void connect() throws SQLException {

      // database + user + options + execFile + tty
      byte [] StartupInfo = new byte[4+4+64+16+64+64+64 + 5];
      int count;

      count = Dbname.length();
      if ( count > 64 ) count = 64;
      //Dbname.getBytes(0,count,StartupInfo,4 + 4);
      System.arraycopy(Dbname.getBytes(),0,StartupInfo,4 + 4,count);

      count = User.length();
      if ( count > 16 ) count = 16;
      //User.getBytes(0,count,StartupInfo,4 + 4 + 64);
      System.arraycopy(User.getBytes(),0,StartupInfo,4 + 4 + 64,count);
      
      // we do nothing with passwords for now
      
      count = Options.length();
      if ( count > 64 ) count = 64;
      //Options.getBytes(0,count,StartupInfo,4 + 4 + 64 + 16);
      System.arraycopy(Options.getBytes(),0,StartupInfo,4 + 4 + 64 + 16,count);
      
// execFile backend specifications are not currently supported
//      count = ExecFile.length();
//      if ( count > 64 ) count = 64;
//      ExecFile.getBytes(0,count,StartupInfo,0);
      
      count = Tty.length();
      if ( count > 64 ) count = 64;
      //Tty.getBytes(0,count,StartupInfo,4 + 4 + 64 + 16 + 64 + 64);
      System.arraycopy(Tty.getBytes(),0,StartupInfo,4 + 4 + 64 + 16 + 64 + 64,count);
      try {
         Socket sock = new Socket(Host,Port);
         sock_in = new DataInputStream(sock.getInputStream());
         sock_out = new DataOutputStream(sock.getOutputStream());
         ByteArrayOutputStream buf = new ByteArrayOutputStream();
         DataOutputStream data = new DataOutputStream(buf);

         count = 64 + 16 + 64 + 64 + 64 + 5 + 4 + 4;
   	 data.writeInt(count);
         data.writeInt(7); // StartupMSG - Username Auth
         data.write(StartupInfo,8,count-8);

         sock_out.write(buf.toByteArray(),0,buf.size());
         sock_out.flush();
      } catch ( Exception e ) {
         throw new SQLException("Server Connection Open Failed: "+e);
      }
      // Send an exec of a null query here to prove operation
      createStatement().executeQuery(" ");
   }

   PGConnection(PGDriver up, String pghost, String pgport, String pgoptions,
                	String pgtty,String dbname,String pguser,
                        String pgpass) throws SQLException {
      parent = up;

      Host = pghost;
      Port = Integer.valueOf(pgport).intValue();
      Options = pgoptions;
      Tty = pgtty;
      Dbname = dbname;
      User = pguser;
      Pass = pgpass;

      connect();
   }
   // createStatement returns an SQL Statement object
   public Statement createStatement() throws SQLException {
      return new PGStatement(this);
   }
   // prepareStatement creates a pre-compiled SQL PreparedStatement object.
   public PreparedStatement prepareStatement(String sql) throws SQLException {
      return new PGPreparedStatement(this,sql);
   }

   // prepareCall create a pre-compiled SQL statement that is
   // a call on a stored procedure.
   public CallableStatement prepareCall(String sql) throws SQLException {
      return new PGCallableStatement(this,sql);
   }

   // Convert the given generic SQL statement to the drivers native SQL.
   public String nativeSQL(String query) throws SQLException {
System.out.println("nativeSQL called");
      return query;
   }

   // If "autoCommit" is true, then all subsequent SQL statements will
   // be executed and committed as individual transactions.  Otherwise
   // (if "autoCommit" is false) then subsequent SQL statements will
   // all be part of the same transaction , which must be explicitly
   // committed with either a "commit" or "rollback" call.
   // By default new connections are initialized with autoCommit "true".
   public void setAutoCommit(boolean enableAutoCommit) throws SQLException {
      if ( (! autoCommit) && enableAutoCommit )
         commit();
      autoCommit = enableAutoCommit;
      if ( ! autoCommit)
         begin();
   }

   public boolean getAutoCommit() throws SQLException
   {
      return false;
   }

   // You can use commit or rollback to commit or abort a transaction.
   // Note that by default all active Statement or ResultSet objects
   // associated with a Connection will be closed when it is committed
   // or aborted.
   protected void begin() throws SQLException {
//      ((PGStatement)createStatement()).exec("begin");
   }

   public void commit() throws SQLException {
//      ((PGStatement)createStatement()).exec("commit");
   }

   public void rollback() throws SQLException {
//      ((PGStatement)createStatement()).exec("rollback");
   }

   // close frees up various state associated with the connection.
   public void close() throws SQLException {
      if ( ! autoCommit )
         commit();
      parent.closeConnection(this);
   }

   // isClosed returns true if the connection is closed, which can
   // occur either due to an explicit call on "close" or due to
   // some fatal error on the connection.
   public boolean isClosed() throws SQLException {
      return false;
   }

   //======================================================================
   // Advanced features:

   // You can obtain a DatabaseMetaData object to get information
   // about the target database.
   public DatabaseMetaData getMetaData() throws SQLException {
      return new PGDatabaseMetaData(this);
   }

   // You can put a connection in read-only mode as a hint to enable
   // database optimizations.  Note that setReadOnly cannot be called
   // while in the middle of a transaction.
   public void setReadOnly(boolean readOnly) throws SQLException {
      ReadOnly=readOnly;
   }
   public boolean isReadOnly() throws SQLException {
      return ReadOnly;
   }

   // The "catalog" selects a sub-space of the target database.
   public void setCatalog(String catalog) throws SQLException {
      Catalog = catalog;
   }
   public String getCatalog() throws SQLException {
      return Catalog;
   }

   // You can call the following method to try to change the transaction
   // isolation level on a newly opened connection, using one of the
   // TRANSACTION_* values.  Use the DatabaseMetaData class to find what
   // isolation levels are supported by the current database.
   // Note that setTransactionIsolation cannot be called while in the
   // middle of a transaction.
   public void setTransactionIsolation(int level) throws SQLException {
      if ( level < TRANSACTION_NONE || level > TRANSACTION_VERSIONING )
         throw new SQLException();
      transactionIsolation = level;
   }
   public int getTransactionIsolation() throws SQLException {
      return transactionIsolation;
   }

   // disableAutoClose can be used on a newly opened connection to to
   // disable the automatic closing of Statements and ResultSets during
   // calls on Commit and Rollback.  This functionality is only available
   // in some databases, and the exact semantics of the objects state
   // vary between databases.  See the DatabaseMetaData class for
   // methods to learn about the consequences of doing disableAutoClose.
   public void disableAutoClose() throws SQLException {
      autoClose = false;
   }

   public void setAutoClose(boolean autoClose) throws SQLException
   {
   }

   public boolean getAutoClose() throws SQLException
   {
      return false;
   }

   // getWarnings will return any warning information related to
   // the current connection.  Note that SQLWarning may be a chain.
   public SQLWarning getWarnings() throws SQLException {
      return warning;
   }
   public void clearWarnings() throws SQLException {
      warning = null;
   }
}

